from .errors import RTError
from .runtime import SymbolTable, Context, RTResult
from .constants import GL_IDENTIFIER
from .lexer import Token
from .nodes import *


class Value:
    def __init__(self):
        self.set_pos()
        self.set_context()

    def set_pos(self, pos_start=None, pos_end=None):
        self.pos_start = pos_start
        self.pos_end = pos_end
        return self

    def set_context(self, context=None):
        self.context = context
        return self

    def added_to(self, other):
        return None, self.illegal_operation(other)

    def subbed_by(self, other):
        return None, self.illegal_operation(other)

    def multed_by(self, other):
        return None, self.illegal_operation(other)

    def dived_by(self, other):
        return None, self.illegal_operation(other)

    def modded_by(self, other):
        return None, self.illegal_operation(other)

    def floordived_by(self, other):
        return None, self.illegal_operation(other)

    def powed_by(self, other):
        return None, self.illegal_operation(other)

    def get_comparison_eq(self, other):
        if not isinstance(other, List):
            return None, Value.illegal_operation(self, other)

        if len(self.elements) != len(other.elements):
            return Number(0).set_context(self.context), None

        for i in range(len(self.elements)):
            result, error = self.elements[i].get_comparison_eq(other.elements[i])
            if error:
                return None, error
            if not result.is_true():
                return Number(0).set_context(self.context), None

        return Number(1).set_context(self.context), None

    def get_comparison_ne(self, other):
        if not isinstance(other, List):
            return None, Value.illegal_operation(self, other)

        result, error = self.get_comparison_eq(other)
        if error:
            return None, error

        if result.is_true():
            return Number(0).set_context(self.context), None
        else:
            return Number(1).set_context(self.context), None

    def get_comparison_lt(self, other):
        return None, self.illegal_operation(other)

    def get_comparison_gt(self, other):
        return None, self.illegal_operation(other)

    def get_comparison_lte(self, other):
        return None, self.illegal_operation(other)

    def get_comparison_gte(self, other):
        return None, self.illegal_operation(other)

    def get_comparison_is(self, other):
        return Number(1 if self is other else 0).set_context(self.context), None

    def anded_by(self, other):
        is_true = self.is_true() and other.is_true()
        return Number(1 if is_true else 0).set_context(self.context), None

    def ored_by(self, other):
        is_true = self.is_true() or other.is_true()
        return Number(1 if is_true else 0).set_context(self.context), None

    def notted(self):
        return None, self.illegal_operation()

    def execute(self, args):
        return RTResult().failure(self.illegal_operation())

    def get_attr(self, name_tok):
        return None, self.illegal_operation()

    def set_attr(self, name_tok, value):
        return None, self.illegal_operation()

    def get_element_at(self, index):
        return None, self.illegal_operation()

    def set_element_at(self, index, value):
        return None, self.illegal_operation()

    def is_true(self):
        return False

    def copy(self):
        raise Exception("No copy method defined")

    def bitted_and_by(self, other):
        return None, self.illegal_operation(other)

    def bitted_or_by(self, other):
        return None, self.illegal_operation(other)

    def bitted_xor_by(self, other):
        return None, self.illegal_operation(other)

    def lshifted_by(self, other):
        return None, self.illegal_operation(other)

    def rshifted_by(self, other):
        return None, self.illegal_operation(other)

    def bitted_not(self):
        return None, self.illegal_operation()

    def illegal_operation(self, other=None):
        if not other:
            other = self
        return RTError(self.pos_start, other.pos_end, "Illegal operation", self.context)


class Number(Value):
    def __init__(self, value):
        super().__init__()
        self.value = value

    def added_to(self, other):
        if isinstance(other, Number):
            return Number(self.value + other.value).set_context(self.context), None
        elif isinstance(other, String):
            return String(str(self.value) + other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def subbed_by(self, other):
        if isinstance(other, Number):
            return Number(self.value - other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def multed_by(self, other):
        if isinstance(other, Number):
            return Number(self.value * other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def dived_by(self, other):
        if isinstance(other, Number):
            if other.value == 0:
                return None, RTError(
                    other.pos_start, other.pos_end, "Division by zero", self.context
                )
            return Number(self.value / other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def modded_by(self, other):
        if isinstance(other, Number):
            if other.value == 0:
                return None, RTError(
                    other.pos_start, other.pos_end, "Division by zero", self.context
                )
            return Number(self.value % other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def floordived_by(self, other):
        if isinstance(other, Number):
            if other.value == 0:
                return None, RTError(
                    other.pos_start, other.pos_end, "Division by zero", self.context
                )
            return Number(self.value // other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_is(self, other):
        return Number(1 if self is other else 0).set_context(self.context), None

    def powed_by(self, other):
        if isinstance(other, Number):
            return Number(self.value**other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_eq(self, other):
        if isinstance(other, Number):
            return (
                Number(int(self.value == other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_ne(self, other):
        if isinstance(other, Number):
            return (
                Number(int(self.value != other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_lt(self, other):
        if isinstance(other, Number):
            return Number(int(self.value < other.value)).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_gt(self, other):
        if isinstance(other, Number):
            return Number(int(self.value > other.value)).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_lte(self, other):
        if isinstance(other, Number):
            return (
                Number(int(self.value <= other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_gte(self, other):
        if isinstance(other, Number):
            return (
                Number(int(self.value >= other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def anded_by(self, other):
        if isinstance(other, Number):
            return (
                Number(int(self.value and other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def ored_by(self, other):
        if isinstance(other, Number):
            return (
                Number(int(self.value or other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def notted(self):
        return Number(1 if self.value == 0 else 0).set_context(self.context), None

    def bitted_and_by(self, other):
        if isinstance(other, Number):
            return (
                Number(int(self.value) & int(other.value)).set_context(self.context),
                None,
            )
        return None, Value.illegal_operation(self, other)

    def bitted_or_by(self, other):
        if isinstance(other, Number):
            return (
                Number(int(self.value) | int(other.value)).set_context(self.context),
                None,
            )
        return None, Value.illegal_operation(self, other)

    def bitted_xor_by(self, other):
        if isinstance(other, Number):
            return (
                Number(int(self.value) ^ int(other.value)).set_context(self.context),
                None,
            )
        return None, Value.illegal_operation(self, other)

    def lshifted_by(self, other):
        if isinstance(other, Number):
            try:
                result = int(self.value) << int(other.value)
                return Number(result).set_context(self.context), None
            except ValueError:
                return None, RTError(
                    other.pos_start, other.pos_end, "Negative shift count", self.context
                )
        return None, Value.illegal_operation(self, other)

    def rshifted_by(self, other):
        if isinstance(other, Number):
            return (
                Number(int(self.value) >> int(other.value)).set_context(self.context),
                None,
            )
        return None, Value.illegal_operation(self, other)

    def bitted_not(self):
        return Number(~int(self.value)).set_context(self.context), None

    def is_true(self):
        return self.value != 0

    def copy(self):
        copy = Number(self.value)
        copy.set_pos(self.pos_start, self.pos_end)
        copy.set_context(self.context)
        return copy

    def __repr__(self):
        return str(self.value)


Number.false = Number(0)
Number.true = Number(1)
Number.null = Number(0)


class String(Value):
    def __init__(self, value):
        super().__init__()
        self.value = value

    def added_to(self, other):
        if isinstance(other, String):
            return String(self.value + other.value).set_context(self.context), None
        elif isinstance(other, Number):
            return String(self.value + str(other.value)).set_context(self.context), None
        else:
            return None, self.illegal_operation(other)

    def get_comparison_eq(self, other):
        if isinstance(other, String):
            return (
                Number(int(self.value == other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_ne(self, other):
        if isinstance(other, String):
            return (
                Number(int(self.value != other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_lt(self, other):
        if isinstance(other, String):
            return Number(int(self.value < other.value)).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_gt(self, other):
        if isinstance(other, String):
            return Number(int(self.value > other.value)).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_lte(self, other):
        if isinstance(other, String):
            return (
                Number(int(self.value <= other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_gte(self, other):
        if isinstance(other, String):
            return (
                Number(int(self.value >= other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def get_element_at(self, index):
        if not isinstance(index, Number):
            return None, RTError(
                self.pos_start,
                self.pos_end,
                "String index must be a Number",
                self.context,
            )

        try:
            val = self.value[int(index.value)]
            return String(val).set_context(self.context), None
        except IndexError:
            return None, RTError(
                self.pos_start,
                self.pos_end,
                f"String index {index.value} out of bounds",
                self.context,
            )

    def is_true(self):
        return len(self.value) > 0

    def copy(self):
        copy = String(self.value)
        copy.set_pos(self.pos_start, self.pos_end)
        copy.set_context(self.context)
        return copy

    def __repr__(self):
        return self.value


class List(Value):
    def __init__(self, elements):
        super().__init__()
        self.elements = elements

    def added_to(self, other):
        if isinstance(other, List):
            new_list = List(self.elements + other.elements)
            new_list.set_context(self.context)
            return new_list, None
        else:
            return None, self.illegal_operation(other)

    def get_element_at(self, index):
        if not isinstance(index, Number):
            return None, RTError(
                self.pos_start,
                self.pos_end,
                "List index must be a Number",
                self.context,
            )

        try:
            element = self.elements[int(index.value)]
            return element, None
        except IndexError:
            return None, RTError(
                self.pos_start,
                self.pos_end,
                f"List index {index.value} out of bounds",
                self.context,
            )

    def set_element_at(self, index, value):
        if not isinstance(index, Number):
            return None, RTError(
                self.pos_start,
                self.pos_end,
                "List index must be a Number",
                self.context,
            )

        try:
            self.elements[int(index.value)] = value
            return value, None
        except IndexError:
            return None, RTError(
                self.pos_start,
                self.pos_end,
                f"List index {index.value} out of bounds",
                self.context,
            )

    def is_true(self):
        return len(self.elements) > 0

    def copy(self):
        copy = List(self.elements[:])
        copy.set_pos(self.pos_start, self.pos_end)
        copy.set_context(self.context)
        return copy

    def __repr__(self):
        return f'[{", ".join([repr(x) for x in self.elements])}]'


class Dict(Value):
    def __init__(self, elements):
        super().__init__()
        self.elements = elements

    def copy(self):
        copy = Dict(self.elements.copy())
        copy.set_pos(self.pos_start, self.pos_end)
        copy.set_context(self.context)
        return copy

    def get_element_at(self, key):
        if not isinstance(key, (Number, String)):
            return None, RTError(
                self.pos_start,
                self.pos_end,
                "Key must be a Number or String",
                self.context,
            )

        val = self.elements.get(key.value)
        if val is None:
            return None, RTError(
                self.pos_start,
                self.pos_end,
                f"Key '{key.value}' not found",
                self.context,
            )
        return val, None

    def set_element_at(self, key, value):
        if not isinstance(key, (Number, String)):
            return None, RTError(
                self.pos_start,
                self.pos_end,
                "Key must be a Number or String",
                self.context,
            )

        self.elements[key.value] = value
        return value, None

    def __repr__(self):
        kv_strings = []
        for key, value in self.elements.items():
            kv_strings.append(f"{repr(key)}: {repr(value)}")
        return f"{{{', '.join(kv_strings)}}}"


class BaseFunction(Value):
    def __init__(self, name):
        super().__init__()
        self.name = name or "<anonymous>"

    def generate_new_context(self):
        new_context = Context(self.name, self.context, self.pos_start)
        new_context.symbol_table = SymbolTable(self.context.symbol_table)
        return new_context

    def check_args(self, arg_names, args):
        res = RTResult()
        if len(args) != len(arg_names):
            return res.failure(
                RTError(
                    self.pos_start,
                    self.pos_end,
                    f"Incorrect argument count for '{self.name}'. Expected {len(arg_names)}, got {len(args)}",
                    self.context,
                )
            )
        return res.success(None)

    def populate_args(self, arg_names, args, new_context):
        for i in range(len(args)):
            arg_name = arg_names[i]
            arg_value = args[i]

            if isinstance(arg_value, BaseFunction):
                pass
            else:
                pass

            new_context.symbol_table.set(arg_name, arg_value)

    def check_and_populate_args(self, arg_names, args, new_context):
        res = RTResult()
        res.register(self.check_args(arg_names, args))
        if res.error:
            return res
        self.populate_args(arg_names, args, new_context)
        return res.success(None)

    def execute(self, args):
        return RTResult().failure(
            RTError(
                self.pos_start,
                self.pos_end,
                "BaseFunction cannot be executed",
                self.context,
            )
        )

    def copy(self):
        raise Exception("Cannot copy a BaseFunction")

    def __repr__(self):
        return f"<function {self.name}>"


class Function(BaseFunction):
    def __init__(self, name, body_node, arg_name_toks, parent_context):
        super().__init__(name)
        self.body_node = body_node
        self.arg_name_toks = arg_name_toks
        self.arg_names = [tok.value for tok in arg_name_toks]
        self.context = parent_context

    def execute(self, args):
        from .interpreter import Interpreter

        res = RTResult()
        interpreter = Interpreter()

        new_context = self.generate_new_context()

        res.register(self.check_and_populate_args(self.arg_names, args, new_context))
        if res.error:
            return res

        value_result = interpreter.visit(self.body_node, new_context)

        if value_result.error:
            return value_result

        if value_result.should_return:
            return res.success(value_result.return_value)

        return res.success(value_result.value or Number.null)

    def copy(self):
        copy = Function(self.name, self.body_node, self.arg_name_toks, self.context)

        copy.set_pos(self.pos_start, self.pos_end)
        copy.set_context(self.context)
        return copy


class Class(BaseFunction):
    def __init__(self, name, superclass, methods):
        super().__init__(name)
        self.superclass = superclass
        self.methods = methods

    def execute(self, args):
        res = RTResult()
        instance = Instance(self)

        fake_init_tok = Token(GL_IDENTIFIER, "init", self.pos_start, self.pos_end)

        init_method, error = self.get_attr(fake_init_tok)

        if error:
            if len(args) > 0:
                return res.failure(
                    RTError(
                        self.pos_start,
                        self.pos_end,
                        f"'{self.name}' does not have an 'init' constructor that accepts {len(args)} arguments",
                        self.context,
                    )
                )
            else:
                return res.success(instance)

        bound_init = init_method.copy().bind_to_instance(instance)
        res.register(bound_init.execute(args))
        if res.error:
            return res

        return res.success(instance)

    def get_attr(self, name_tok):
        method_name = name_tok.value
        method = self.methods.get(method_name)

        if method:
            return (
                method.copy()
                .set_context(self.context)
                .set_pos(name_tok.pos_start, name_tok.pos_end),
                None,
            )

        if self.superclass:
            return self.superclass.get_attr(name_tok)

        return None, RTError(
            name_tok.pos_start,
            name_tok.pos_end,
            f"Class '{self.name}' has no method '{method_name}'",
            self.context,
        )

    def copy(self):
        copy = Class(self.name, self.superclass, self.methods)
        copy.set_context(self.context)
        copy.set_pos(self.pos_start, self.pos_end)
        return copy

    def __repr__(self):
        return f"<class {self.name}>"


class Instance(Value):
    def __init__(self, class_ref):
        super().__init__()
        self.class_ref = class_ref
        self.symbol_table = SymbolTable()

    def get_attr(self, name_tok):
        name = name_tok.value

        value = self.symbol_table.get(name)
        if value:
            return value, None

        method, error = self.class_ref.get_attr(name_tok)
        if error:
            return None, error

        bound_method = method.copy().bind_to_instance(self)
        return bound_method, None

    def set_attr(self, name_tok, value):
        name = name_tok.value
        self.symbol_table.set(name, value)
        return value, None

    def copy(self):
        copy = Instance(self.class_ref)
        copy.symbol_table = self.symbol_table
        copy.set_context(self.context)
        copy.set_pos(self.pos_start, self.pos_end)
        return copy

    def __repr__(self):
        return f"<{self.class_ref.name} instance>"


class BoundMethod(BaseFunction):
    def __init__(self, name, function_to_bind, instance):
        super().__init__(name)
        self.function_to_bind = function_to_bind
        self.instance = instance
        self.context = function_to_bind.context
        self.set_pos(function_to_bind.pos_start, function_to_bind.pos_end)

    def execute(self, args):
        from .interpreter import Interpreter

        res = RTResult()
        interpreter = Interpreter()

        new_context = self.function_to_bind.generate_new_context()
        original_arg_names = self.function_to_bind.arg_names

        if len(original_arg_names) > 0 and original_arg_names[0] == "SELF":
            new_context.symbol_table.set("SELF", self.instance)

        if len(original_arg_names) == 0 or original_arg_names[0] != "SELF":
            return res.failure(
                RTError(
                    self.function_to_bind.pos_start,
                    self.function_to_bind.pos_end,
                    f"Method '{self.name}' must have 'SELF' as its first argument",
                    self.context,
                )
            )

        expected_arg_names = original_arg_names[1:]

        res.register(
            self.function_to_bind.check_and_populate_args(
                expected_arg_names, args, new_context
            )
        )

        if res.error:
            return res

        value_result = interpreter.visit(self.function_to_bind.body_node, new_context)

        if value_result.error:
            return value_result

        if value_result.should_return:
            return res.success(value_result.return_value)

        return res.success(value_result.value or Number.null)

    def copy(self):
        return (
            BoundMethod(self.name, self.function_to_bind.copy(), self.instance)
            .set_context(self.context)
            .set_pos(self.pos_start, self.pos_end)
        )


class BuiltInFunction(BaseFunction):
    def __init__(self, name):
        super().__init__(name)

    def execute(self, args):
        res = RTResult()

        if self.name == "INPUT":
            if len(args) > 1:
                return res.failure(
                    RTError(
                        self.pos_start,
                        self.pos_end,
                        "INPUT takes at most 1 argument",
                        self.context,
                    )
                )

            prompt = ""
            if len(args) == 1:
                prompt = str(args[0].value)

            try:
                text = input(prompt)
            except EOFError:
                text = ""
            return res.success(String(text))

        elif self.name == "STR":
            res.register(self.check_args(["value"], args))
            if res.error:
                return res

            val = args[0]
            if isinstance(val, String):
                return res.success(String(val.value))
            else:
                return res.success(String(str(val)))

        elif self.name == "INT":
            res.register(self.check_args(["value"], args))
            if res.error:
                return res
            arg = args[0]

            if not isinstance(arg, (Number, String)):
                return res.failure(
                    RTError(
                        self.pos_start,
                        self.pos_end,
                        f"Argument for INT must be a Number or String, got {type(arg).__name__}",
                        self.context,
                    )
                )

            try:
                val = int(float(arg.value))
                return res.success(Number(val))
            except ValueError:
                return res.failure(
                    RTError(
                        self.pos_start,
                        self.pos_end,
                        f"Cannot convert '{arg.value}' to INT",
                        self.context,
                    )
                )

        elif self.name == "FLOAT":
            res.register(self.check_args(["value"], args))
            if res.error:
                return res
            arg = args[0]

            if not isinstance(arg, (Number, String)):
                return res.failure(
                    RTError(
                        self.pos_start,
                        self.pos_end,
                        f"Argument for FLOAT must be a Number or String, got {type(arg).__name__}",
                        self.context,
                    )
                )

            try:
                val = float(arg.value)
                return res.success(Number(val))
            except ValueError:
                return res.failure(
                    RTError(
                        self.pos_start,
                        self.pos_end,
                        f"Cannot convert '{arg.value}' to FLOAT",
                        self.context,
                    )
                )

        elif self.name == "BOOL":
            res.register(self.check_args(["value"], args))
            if res.error:
                return res

            is_true = args[0].is_true()
            return res.success(Number.true if is_true else Number.false)

        elif self.name == "LEN":
            res.register(self.check_args(["value"], args))
            if res.error:
                return res

            arg = args[0]

            if isinstance(arg, String):
                return res.success(Number(len(arg.value)))
            elif isinstance(arg, List):
                return res.success(Number(len(arg.elements)))
            elif isinstance(arg, Dict):
                return res.success(Number(len(arg.elements)))
            elif isinstance(arg, Number):
                return res.success(Number(len(str(arg.value))))
            elif isinstance(arg, (Function, BuiltInFunction, Class)):
                return res.success(Number(1))

            return res.success(Number(0))

        return res.failure(
            RTError(
                self.pos_start,
                self.pos_end,
                f"Built-in function '{self.name}' is not defined.",
                self.context,
            )
        )

    def copy(self):
        copy = BuiltInFunction(self.name)
        copy.set_context(self.context)
        copy.set_pos(self.pos_start, self.pos_end)
        return copy

    def __repr__(self):
        return f"<built-in function {self.name}>"


def bind_to_instance(self, instance):
    return BoundMethod(self.name, self, instance)


Function.bind_to_instance = bind_to_instance



from .runtime import RTResult, Context, SymbolTable
from .values import Number, String, Function, Class, List, Dict
from .nodes import *
from .errors import RTError
from .constants import *


class Interpreter:
    def visit(self, node, context):
        method_name = f"visit_{type(node).__name__}"
        method = getattr(self, method_name, self.no_visit_method)
        result = method(node, context)

        if isinstance(result, RTResult) and (
            result.should_return
            or result.should_break
            or result.should_continue
            or result.error
        ):
            return result

        return result

    def no_visit_method(self, node, context):
        raise Exception(f"No visit_{type(node).__name__} method defined")

    def visit_StatementListNode(self, node, context):
        res = RTResult()
        last_value = Number.null

        for statement_node in node.statement_nodes:
            last_value = res.register(self.visit(statement_node, context))

            if (
                res.error
                or res.should_return
                or res.should_break
                or res.should_continue
            ):
                return res

        return res.success(last_value)

    def visit_NumberNode(self, node, context):
        return RTResult().success(
            Number(node.tok.value)
            .set_context(context)
            .set_pos(node.pos_start, node.pos_end)
        )

    def visit_StringNode(self, node, context):
        return RTResult().success(
            String(node.tok.value)
            .set_context(context)
            .set_pos(node.pos_start, node.pos_end)
        )

    def visit_ListNode(self, node, context):
        res = RTResult()
        elements = []

        for element_node in node.element_nodes:
            elements.append(res.register(self.visit(element_node, context)))
            if res.error:
                return res

        return res.success(
            List(elements).set_context(context).set_pos(node.pos_start, node.pos_end)
        )

    def visit_DictNode(self, node, context):
        res = RTResult()
        elements = {}

        for key_node, value_node in node.key_value_pairs:
            key = res.register(self.visit(key_node, context))
            if res.error:
                return res

            value = res.register(self.visit(value_node, context))
            if res.error:
                return res

            if isinstance(key, (Number, String)):
                elements[key.value] = value
            else:
                return res.failure(
                    RTError(
                        key_node.pos_start,
                        key_node.pos_end,
                        "Dictionary key must be a Number or String",
                        context,
                    )
                )

        return res.success(
            Dict(elements).set_context(context).set_pos(node.pos_start, node.pos_end)
        )

    def visit_MultiVarAssignNode(self, node, context):
        res = RTResult()

        list_val = res.register(self.visit(node.value_node, context))
        if res.error:
            return res

        if not isinstance(list_val, List):
            return res.failure(
                RTError(
                    node.pos_start,
                    node.pos_end,
                    f"Cannot unpack type '{type(list_val).__name__}' (expected List)",
                    context,
                )
            )

        if len(node.var_name_toks) != len(list_val.elements):
            return res.failure(
                RTError(
                    node.pos_start,
                    node.pos_end,
                    f"ValueError: too many/not enough values to unpack (expected {len(node.var_name_toks)}, got {len(list_val.elements)})",
                    context,
                )
            )

        for i, var_name_tok in enumerate(node.var_name_toks):
            var_name = var_name_tok.value

            if var_name in context.symbol_table.finals:
                return res.failure(
                    RTError(
                        var_name_tok.pos_start,
                        var_name_tok.pos_end,
                        f"Cannot reassign constant '{var_name}'",
                        context,
                    )
                )

            val = list_val.elements[i]
            context.symbol_table.set(var_name, val)

        return res.success(list_val)

    def visit_ListCompNode(self, node, context):
        res = RTResult()
        output_list = []

        if node.var_name_tok.value in context.symbol_table.finals:
            return res.failure(
                RTError(
                    node.var_name_tok.pos_start,
                    node.var_name_tok.pos_end,
                    f"Cannot use constant '{node.var_name_tok.value}' as comprehension variable",
                    context,
                )
            )

        iterable_val = res.register(self.visit(node.iterable_node, context))
        if res.error:
            return res

        if not isinstance(iterable_val, List):
            return res.failure(
                RTError(
                    node.iterable_node.pos_start,
                    node.iterable_node.pos_end,
                    "Expected a list",
                    context,
                )
            )

        for element in iterable_val.elements:
            context.symbol_table.set(node.var_name_tok.value, element)

            value = res.register(self.visit(node.output_expr_node, context))
            if res.error:
                return res

            output_list.append(value)

        return res.success(
            List(output_list).set_context(context).set_pos(node.pos_start, node.pos_end)
        )

    def visit_SliceAccessNode(self, node, context):
        res = RTResult()

        obj = res.register(self.visit(node.node_to_slice, context))
        if res.error:
            return res

        start_val = res.register(self.visit(node.start_node, context))
        if res.error:
            return res

        end_val = None
        if node.end_node:
            end_val = res.register(self.visit(node.end_node, context))
            if res.error:
                return res

        if isinstance(obj, (List, String)):
            if not isinstance(start_val, Number):
                return res.failure(
                    RTError(
                        node.start_node.pos_start,
                        node.start_node.pos_end,
                        "Start index must be a number",
                        context,
                    )
                )

            start_idx = int(start_val.value)
            end_idx = None

            if end_val:
                if not isinstance(end_val, Number):
                    return res.failure(
                        RTError(
                            node.end_node.pos_start,
                            node.end_node.pos_end,
                            "End index must be a number",
                            context,
                        )
                    )
                end_idx = int(end_val.value)

            if isinstance(obj, List):
                new_elements = obj.elements[start_idx:end_idx]
                return res.success(
                    List(new_elements)
                    .set_context(context)
                    .set_pos(node.pos_start, node.pos_end)
                )

            elif isinstance(obj, String):
                new_str = obj.value[start_idx:end_idx]
                return res.success(
                    String(new_str)
                    .set_context(context)
                    .set_pos(node.pos_start, node.pos_end)
                )

        return res.failure(
            RTError(
                node.pos_start,
                node.pos_end,
                f"Type {type(obj).__name__} is not sliceable",
                context,
            )
        )

    def visit_VarAccessNode(self, node, context):
        res = RTResult()
        var_name = node.var_name_tok.value
        value = context.symbol_table.get(var_name)

        if not value:
            return res.failure(
                RTError(
                    node.pos_start,
                    node.pos_end,
                    f"'{var_name}' is not defined",
                    context,
                )
            )

        value = value.set_pos(node.pos_start, node.pos_end)

        return res.success(value)

    def visit_VarAssignNode(self, node, context):
        res = RTResult()
        var_name = node.var_name_tok.value
        value = res.register(self.visit(node.value_node, context))
        if res.error:
            return res

        context.symbol_table.set(var_name, value)
        return res.success(value)

    def visit_PrintNode(self, node, context):
        res = RTResult()
        value = res.register(self.visit(node.node_to_print, context))
        if res.error:
            return res

        print(value)
        return res.success(Number.null)

    def visit_IfNode(self, node, context):
        res = RTResult()

        for condition, body in node.cases:
            condition_value = res.register(self.visit(condition, context))
            if res.error:
                return res

            if condition_value.is_true():
                expr_value = res.register(self.visit(body, context))
                if res.error:
                    return res

                if res.should_return or res.should_break or res.should_continue:
                    return res

                return res.success(expr_value)

        if node.else_case:
            expr_value = res.register(self.visit(node.else_case, context))
            if res.error:
                return res

            if res.should_return or res.should_break or res.should_continue:
                return res

            return res.success(expr_value)

        return res.success(Number.null)

    def visit_ForNode(self, node, context):
        res = RTResult()

        if node.var_name_tok.value in context.symbol_table.finals:
            return res.failure(
                RTError(
                    node.var_name_tok.pos_start,
                    node.var_name_tok.pos_end,
                    f"Cannot use constant '{node.var_name_tok.value}' as for loop variable",
                    context,
                )
            )

        iterable_value = res.register(self.visit(node.iterable_node, context))
        if res.error:
            return res

        if not isinstance(iterable_value, List):
            return res.failure(
                RTError(
                    node.iterable_node.pos_start,
                    node.iterable_node.pos_end,
                    "Iterable must be a List",
                    context,
                )
            )

        for element in iterable_value.elements:
            context.symbol_table.set(node.var_name_tok.value, element)

            value = res.register(self.visit(node.body_node, context))
            if res.error:
                return res

            if res.should_continue:
                res.should_continue = False
                continue

            if res.should_break:
                res.should_break = False
                break

            if res.should_return:
                return res

        return res.success(Number.null)

    def visit_WhileNode(self, node, context):
        res = RTResult()

        while True:
            condition_value = res.register(self.visit(node.condition_node, context))
            if res.error:
                return res

            if not condition_value.is_true():
                break

            value = res.register(self.visit(node.body_node, context))
            if res.error:
                return res

            if res.should_continue:
                res.should_continue = False
                continue

            if res.should_break:
                res.should_break = False
                break

            if res.should_return:
                return res

        return res.success(Number.null)

    def visit_BreakNode(self, node, context):
        return RTResult().success_break()

    def visit_ContinueNode(self, node, context):
        return RTResult().success_continue()

    def visit_FunDefNode(self, node, context):
        res = RTResult()

        func_name = node.var_name_tok.value if node.var_name_tok else None
        body_node = node.body_node

        func = Function(func_name, body_node, node.arg_name_toks, context)

        if func_name:
            context.symbol_table.set(func_name, func)

        return res.success(
            func.set_pos(node.pos_start, node.pos_end).set_context(context)
        )

    def visit_CallNode(self, node, context):
        res = RTResult()
        args = []

        value_to_call = res.register(self.visit(node.node_to_call, context))
        if res.error:
            return res

        value_to_call = value_to_call.copy().set_pos(node.pos_start, node.pos_end)

        for arg_node in node.arg_nodes:
            args.append(res.register(self.visit(arg_node, context)))
            if res.error:
                return res

        return_value = res.register(value_to_call.execute(args))
        if res.error:
            return res

        return res.success(return_value)

    def visit_ReturnNode(self, node, context):
        res = RTResult()

        value = res.register(self.visit(node.node_to_return, context))
        if res.error:
            return res

        return res.success_return(value)

    def visit_ClassNode(self, node, context):
        res = RTResult()
        class_name = node.class_name_tok.value

        superclass = None
        if node.superclass_node:
            superclass = res.register(self.visit(node.superclass_node, context))
            if res.error:
                return res
            if not isinstance(superclass, Class):
                return res.failure(
                    RTError(
                        node.superclass_node.pos_start,
                        node.superclass_node.pos_end,
                        "A class can only inherit from another class",
                        context,
                    )
                )

        methods = {}
        for method_node in node.method_nodes:
            method_name = method_node.var_name_tok.value
            method_value = Function(
                method_name, method_node.body_node, method_node.arg_name_toks, context
            ).set_pos(method_node.pos_start, method_node.pos_end)
            methods[method_name] = method_value

        class_value = Class(class_name, superclass, methods)
        class_value.set_context(context).set_pos(node.pos_start, node.pos_end)

        context.symbol_table.set(class_name, class_value)
        return res.success(class_value)

    def visit_NewInstanceNode(self, node, context):
        res = RTResult()

        class_name = node.class_name_tok.value
        class_value = context.symbol_table.get(class_name)

        if not class_value:
            return res.failure(
                RTError(
                    node.pos_start,
                    node.pos_end,
                    f"Class '{class_name}' is not defined",
                    context,
                )
            )

        if not isinstance(class_value, Class):
            return res.failure(
                RTError(
                    node.pos_start,
                    node.pos_end,
                    f"'{class_name}' is not a class",
                    context,
                )
            )

        args = []
        for arg_node in node.arg_nodes:
            args.append(res.register(self.visit(arg_node, context)))
            if res.error:
                return res

        instance = res.register(class_value.execute(args))
        if res.error:
            return res

        return res.success(instance.set_pos(node.pos_start, node.pos_end))

    def visit_GetAttrNode(self, node, context):
        res = RTResult()

        object = res.register(self.visit(node.object_node, context))
        if res.error:
            return res

        value, error = object.get_attr(node.attr_name_tok)
        if error:
            return res.failure(error)

        return res.success(value.set_pos(node.pos_start, node.pos_end))

    def visit_SetAttrNode(self, node, context):
        res = RTResult()

        object = res.register(self.visit(node.object_node, context))
        if res.error:
            return res

        value = res.register(self.visit(node.value_node, context))
        if res.error:
            return res

        new_value, error = object.set_attr(node.attr_name_tok, value)
        if error:
            return res.failure(error)

        return res.success(new_value)

    def visit_ListAccessNode(self, node, context):
        res = RTResult()

        list_val = res.register(self.visit(node.list_node, context))
        if res.error:
            return res

        index_val = res.register(self.visit(node.index_node, context))
        if res.error:
            return res

        element, error = list_val.get_element_at(index_val)
        if error:
            return res.failure(error)

        return res.success(element.copy().set_pos(node.pos_start, node.pos_end))

    def visit_ListSetNode(self, node, context):
        res = RTResult()

        list_val = res.register(self.visit(node.list_node, context))
        if res.error:
            return res

        index_val = res.register(self.visit(node.index_node, context))
        if res.error:
            return res

        value_to_set = res.register(self.visit(node.value_node, context))
        if res.error:
            return res

        new_value, error = list_val.set_element_at(index_val, value_to_set)
        if error:
            return res.failure(error)

        return res.success(new_value)

    def visit_BinOpNode(self, node, context):
        res = RTResult()
        left = res.register(self.visit(node.left_node, context))
        if res.error:
            return res
        right = res.register(self.visit(node.right_node, context))
        if res.error:
            return res

        if node.op_tok.type == GL_PLUS:
            result, error = left.added_to(right)
        elif node.op_tok.type == GL_MINUS:
            result, error = left.subbed_by(right)
        elif node.op_tok.type == GL_MUL:
            result, error = left.multed_by(right)
        elif node.op_tok.type == GL_DIV:
            result, error = left.dived_by(right)
        elif node.op_tok.type == GL_MOD:
            result, error = left.modded_by(right)
        elif node.op_tok.type == GL_FLOORDIV:
            result, error = left.floordived_by(right)
        elif node.op_tok.type == GL_POW:
            result, error = left.powed_by(right)
        elif node.op_tok.type == GL_EE:
            result, error = left.get_comparison_eq(right)
        elif node.op_tok.type == GL_NE:
            result, error = left.get_comparison_ne(right)
        elif node.op_tok.type == GL_LT:
            result, error = left.get_comparison_lt(right)
        elif node.op_tok.type == GL_GT:
            result, error = left.get_comparison_gt(right)
        elif node.op_tok.type == GL_LTE:
            result, error = left.get_comparison_lte(right)
        elif node.op_tok.type == GL_GTE:
            result, error = left.get_comparison_gte(right)
        elif node.op_tok.matches(GL_KEYWORD, "AND") or node.op_tok.matches(
            GL_KEYWORD, "and"
        ):
            result, error = left.anded_by(right)
        elif node.op_tok.matches(GL_KEYWORD, "OR") or node.op_tok.matches(
            GL_KEYWORD, "or"
        ):
            result, error = left.ored_by(right)
        elif node.op_tok.matches(GL_KEYWORD, "IS") or node.op_tok.matches(
            GL_KEYWORD, "is"
        ):
            result, error = left.get_comparison_is(right)

        elif node.op_tok.type == GL_BIT_AND:
            result, error = left.bitted_and_by(right)
        elif node.op_tok.type == GL_BIT_OR:
            result, error = left.bitted_or_by(right)
        elif node.op_tok.type == GL_BIT_XOR:
            result, error = left.bitted_xor_by(right)
        elif node.op_tok.type == GL_LSHIFT:
            result, error = left.lshifted_by(right)
        elif node.op_tok.type == GL_RSHIFT:
            result, error = left.rshifted_by(right)

        if error:
            return res.failure(error)
        else:
            return res.success(result.set_pos(node.pos_start, node.pos_end))

    def visit_UnaryOpNode(self, node, context):
        res = RTResult()

        if node.op_tok.type in (GL_PLUSPLUS, GL_MINUSMINUS):
            target_node = node.node

            if isinstance(target_node, VarAccessNode):
                var_name = target_node.var_name_tok.value
                value = context.symbol_table.get(var_name)
                if not value:
                    return res.failure(
                        RTError(
                            target_node.pos_start,
                            target_node.pos_end,
                            f"'{var_name}' is not defined",
                            context,
                        )
                    )

            elif isinstance(target_node, GetAttrNode):
                obj = res.register(self.visit(target_node.object_node, context))
                if res.error:
                    return res
                value, error = obj.get_attr(target_node.attr_name_tok)
                if error:
                    return res.failure(error)

            elif isinstance(target_node, ListAccessNode):
                list_val = res.register(self.visit(target_node.list_node, context))
                if res.error:
                    return res
                index_val = res.register(self.visit(target_node.index_node, context))
                if res.error:
                    return res
                value, error = list_val.get_element_at(index_val)
                if error:
                    return res.failure(error)

            else:
                return res.failure(
                    RTError(
                        target_node.pos_start,
                        target_node.pos_end,
                        "Invalid target for increment/decrement",
                        context,
                    )
                )

            if not isinstance(value, Number):
                return res.failure(
                    RTError(
                        target_node.pos_start,
                        target_node.pos_end,
                        "Operand must be a number",
                        context,
                    )
                )

            if node.op_tok.type == GL_PLUSPLUS:
                new_value, error = value.added_to(Number(1))
            else:
                new_value, error = value.subbed_by(Number(1))

            if error:
                return res.failure(error)

            if isinstance(target_node, VarAccessNode):
                err = context.symbol_table.update(var_name, new_value)
                if err:
                    return res.failure(
                        RTError(
                            target_node.pos_start, target_node.pos_end, err, context
                        )
                    )

            if isinstance(target_node, VarAccessNode):
                context.symbol_table.set(var_name, new_value)

            elif isinstance(target_node, GetAttrNode):
                _, error = obj.set_attr(target_node.attr_name_tok, new_value)
                if error:
                    return res.failure(error)

            elif isinstance(target_node, ListAccessNode):
                _, error = list_val.set_element_at(index_val, new_value)
                if error:
                    return res.failure(error)

            return res.success(new_value.copy().set_pos(node.pos_start, node.pos_end))

        number = res.register(self.visit(node.node, context))
        if res.error:
            return res

        number = number.copy()

        error = None
        if node.op_tok.type == GL_MINUS:
            if isinstance(number, Number):
                number, error = number.multed_by(Number(-1))
            else:
                error = RTError(
                    node.pos_start,
                    node.pos_end,
                    "Unary '-' can only be applied to numbers",
                    context,
                )
        elif node.op_tok.matches(GL_KEYWORD, "NOT"):
            number, error = number.notted()

        elif node.op_tok.type == GL_BIT_NOT:
            number, error = number.bitted_not()

        if error:
            return res.failure(error)
        else:
            return res.success(number.set_pos(node.pos_start, node.pos_end))

    def visit_PostOpNode(self, node, context):
        res = RTResult()
        target_node = node.node

        if isinstance(target_node, VarAccessNode):
            var_name = target_node.var_name_tok.value
            old_value = context.symbol_table.get(var_name)
            if not old_value:
                return res.failure(
                    RTError(
                        target_node.pos_start,
                        target_node.pos_end,
                        f"'{var_name}' is not defined",
                        context,
                    )
                )

        elif isinstance(target_node, GetAttrNode):
            obj = res.register(self.visit(target_node.object_node, context))
            if res.error:
                return res
            old_value, error = obj.get_attr(target_node.attr_name_tok)
            if error:
                return res.failure(error)

        elif isinstance(target_node, ListAccessNode):
            list_val = res.register(self.visit(target_node.list_node, context))
            if res.error:
                return res
            index_val = res.register(self.visit(target_node.index_node, context))
            if res.error:
                return res
            old_value, error = list_val.get_element_at(index_val)
            if error:
                return res.failure(error)

        else:
            return res.failure(
                RTError(
                    target_node.pos_start,
                    target_node.pos_end,
                    "Invalid target for increment/decrement",
                    context,
                )
            )

        if not isinstance(old_value, Number):
            return res.failure(
                RTError(
                    target_node.pos_start,
                    target_node.pos_end,
                    "Operand must be a number",
                    context,
                )
            )

        if node.op_tok.type == GL_PLUSPLUS:
            new_value, error = old_value.added_to(Number(1))
        else:
            new_value, error = old_value.subbed_by(Number(1))

        if error:
            return res.failure(error)

        if isinstance(target_node, VarAccessNode):
            err = context.symbol_table.update(var_name, new_value)
            if err:
                return res.failure(
                    RTError(target_node.pos_start, target_node.pos_end, err, context)
                )

        elif isinstance(target_node, GetAttrNode):
            _, error = obj.set_attr(target_node.attr_name_tok, new_value)
            if error:
                return res.failure(error)

        elif isinstance(target_node, ListAccessNode):
            _, error = list_val.set_element_at(index_val, new_value)
            if error:
                return res.failure(error)

        return res.success(old_value.copy().set_pos(node.pos_start, node.pos_end))

    def visit_TryCatchNode(self, node, context):
        res = RTResult()

        try_res = self.visit(node.try_body_node, context)

        if try_res.error:
            if node.catch_body_node:
                catch_context = Context("CATCH", context, node.pos_start)
                catch_context.symbol_table = SymbolTable(context.symbol_table)

                if node.catch_var_node:
                    error_msg = try_res.error.details
                    val_to_assign = getattr(try_res.error, "thrown_value", None)
                    if val_to_assign is None:
                        val_to_assign = String(error_msg)
                    catch_context.symbol_table.set(
                        node.catch_var_node.value, val_to_assign
                    )

                catch_res = res.register(
                    self.visit(node.catch_body_node, catch_context)
                )

                if res.error:
                    if node.finally_body_node:
                        fin_res = self.visit(node.finally_body_node, context)
                        if fin_res.error:
                            return res.failure(fin_res.error)

                    return res

            else:
                if node.finally_body_node:
                    fin_res = self.visit(node.finally_body_node, context)
                    if fin_res.error:
                        return fin_res
                return try_res
        else:
            res.register(try_res)
            if res.should_return or res.should_break or res.should_continue:
                if node.finally_body_node:
                    self.visit(node.finally_body_node, context)
                return res

        if node.finally_body_node:
            fin_res = res.register(self.visit(node.finally_body_node, context))
            if res.error:
                return res

        return res.success(Number.null)

    def visit_ThrowNode(self, node, context):
        res = RTResult()

        value = res.register(self.visit(node.node_to_throw, context))
        if res.error:
            return res

        error_message = str(value)

        return res.failure(
            RTError(
                node.pos_start, node.pos_end, error_message, context, thrown_value=value
            )
        )

    def visit_FinalVarAssignNode(self, node, context):
        res = RTResult()
        var_name = node.var_name_tok.value

        if var_name in context.symbol_table.symbols:
            return res.failure(
                RTError(
                    node.pos_start,
                    node.pos_end,
                    f"Variable '{var_name}' is already defined",
                    context,
                )
            )

        value = res.register(self.visit(node.value_node, context))
        if res.error:
            return res

        context.symbol_table.set(var_name, value, as_final=True)
        return res.success(value)

    def visit_VarAssignNode(self, node, context):
        res = RTResult()
        var_name = node.var_name_tok.value

        if var_name in context.symbol_table.finals:
            return res.failure(
                RTError(
                    node.pos_start,
                    node.pos_end,
                    f"Cannot reassign constant '{var_name}'",
                    context,
                )
            )

        value = res.register(self.visit(node.value_node, context))
        if res.error:
            return res

        context.symbol_table.set(var_name, value)
        return res.success(value)

    def visit_SwitchNode(self, node, context):
        res = RTResult()

        switch_val = res.register(self.visit(node.switch_value_node, context))
        if res.error:
            return res

        for case_conditions, body_node in node.cases:
            should_execute = False

            for cond_node in case_conditions:
                case_val = res.register(self.visit(cond_node, context))
                if res.error:
                    return res

                is_eq, error = switch_val.get_comparison_eq(case_val)
                if error:
                    return res.failure(error)

                if is_eq.is_true():
                    should_execute = True
                    break

            if should_execute:
                val = res.register(self.visit(body_node, context))
                if (
                    res.error
                    or res.should_return
                    or res.should_break
                    or res.should_continue
                ):
                    return res
                return res.success(val)

        if node.default_case:
            val = res.register(self.visit(node.default_case, context))
            if (
                res.error
                or res.should_return
                or res.should_break
                or res.should_continue
            ):
                return res
            return res.success(val)

        return res.success(Number.null)



from .errors import RTError


class SymbolTable:
    def __init__(self, parent=None):
        self.symbols = {}
        self.finals = []
        self.parent = parent

    def get(self, name):
        value = self.symbols.get(name, None)
        if value == None and self.parent:
            return self.parent.get(name)
        return value

    def set(self, name, value, as_final=False):
        self.symbols[name] = value
        if as_final:
            self.finals.append(name)

    def remove(self, name):
        del self.symbols[name]
        if name in self.finals:
            self.finals.remove(name)

    def update(self, name, value):
        if name in self.symbols:
            if name in self.finals:
                return f"Cannot reassign constant '{name}'"

            self.symbols[name] = value
            return None

        if self.parent:
            return self.parent.update(name, value)

        return f"'{name}' is not defined"


class Context:
    def __init__(self, display_name, parent=None, parent_entry_pos=None):
        self.display_name = display_name
        self.parent = parent
        self.parent_entry_pos = parent_entry_pos
        self.symbol_table = None


class RTResult:
    def __init__(self):
        self.value = None
        self.error = None
        self.return_value = None
        self.should_return = False
        self.should_break = False
        self.should_continue = False

    def register(self, res):
        if res.error:
            self.error = res.error
        if res.should_return:
            self.return_value = res.return_value
            self.should_return = True
        if res.should_break:
            self.should_break = True
        if res.should_continue:
            self.should_continue = True
        return res.value

    def success(self, value):
        self.value = value
        return self

    def success_return(self, value):
        self.return_value = value
        self.should_return = True
        return self

    def success_break(self):
        self.should_break = True
        return self

    def success_continue(self):
        self.should_continue = True
        return self

    def failure(self, error):
        self.error = error
        return self
